{-# OPTIONS_GHC -fglasgow-exts #-}

module Polynomial where {

import qualified Nat;
import qualified List;
import qualified Term;
import qualified Show;
import qualified Option;
import qualified Product_Type;
import qualified Util;
import qualified ErrorMonad;
import qualified RedPair;
import qualified UsableRules;
import qualified HOL;

mul :: Nat.Nat -> Util.Cint -> Util.Cint;
mul n (Util.Neg m) = Util.Neg (Nat.times_nat n m);
mul n (Util.Pos m) = Util.Pos (Nat.times_nat n m);

data Poly a = Poly [(Nat.Nat, a)] Util.Cint;

data RedPair a = Polo [(a, (Nat.Nat, [Nat.Nat]))]
  | NegPolo [(a, (Util.Cint, [Nat.Nat]))];

addI :: Util.Cint -> Util.Cint -> Util.Cint;
addI (Util.Neg x) (Util.Neg y) = Util.Neg (Nat.plus_nat x y);
addI (Util.Pos x) (Util.Pos y) = Util.Pos (Nat.plus_nat x y);
addI (Util.Neg x) (Util.Pos y) =
  (if Nat.less_eq_nat x y then Util.Pos (Nat.minus_nat y x)
    else Util.Neg (Nat.minus_nat x y));
addI (Util.Pos x) (Util.Neg y) =
  (if Nat.less_eq_nat y x then Util.Pos (Nat.minus_nat x y)
    else Util.Neg (Nat.minus_nat y x));

polyMult :: forall a. Nat.Nat -> Polynomial.Poly a -> Polynomial.Poly a;
polyMult n (Polynomial.Poly vp c) =
  Polynomial.Poly (List.mapa (\ cx -> (Nat.times_nat n (fst cx), snd cx)) vp)
    (Polynomial.mul n c);

lookupVar ::
  forall a. (Eq a) => a -> [(Nat.Nat, a)] -> Maybe (Nat.Nat, [(Nat.Nat, a)]);
lookupVar x [] = Nothing;
lookupVar x ((c, y) : vp) =
  (if x == y then Just (c, vp)
    else (case Polynomial.lookupVar x vp of {
           Nothing -> Nothing;
           Just a -> let {
                       (d, vpp) = a;
                     } in Just (d, (c, y) : vpp);
         }));

sumPolyVars ::
  forall a. (Eq a) => [(Nat.Nat, a)] -> [(Nat.Nat, a)] -> [(Nat.Nat, a)];
sumPolyVars [] vp2 = vp2;
sumPolyVars ((c, x) : vp1) vp2 =
  (case Polynomial.lookupVar x vp2 of {
    Nothing -> (c, x) : Polynomial.sumPolyVars vp1 vp2;
    Just a ->
      let {
        (d, vpp2) = a;
      } in (Nat.plus_nat c d, x) : Polynomial.sumPolyVars vp1 vpp2;
  });

cleanPolyVars :: forall a. [(Nat.Nat, a)] -> [(Nat.Nat, a)];
cleanPolyVars [] = [];
cleanPolyVars ((c, x) : vp) =
  List.append (if c == Nat.Zero_nat then [] else [(c, x)])
    (Polynomial.cleanPolyVars vp);

polySum ::
  forall a.
    (Eq a) => Polynomial.Poly a -> Polynomial.Poly a -> Polynomial.Poly a;
polySum (Polynomial.Poly vp1 c) (Polynomial.Poly vp2 d) =
  Polynomial.Poly (Polynomial.cleanPolyVars (Polynomial.sumPolyVars vp1 vp2))
    (Polynomial.addI c d);

poly_case ::
  forall v t. ([(Nat.Nat, v)] -> Util.Cint -> t) -> Polynomial.Poly v -> t;
poly_case f1 (Polynomial.Poly list cint) = f1 list cint;

pleft ::
  forall a b.
    (Eq b) => (a -> Nat.Nat -> ([Nat.Nat], Util.Cint)) ->
                Term.Term a b -> Polynomial.Poly b;
pleft alpha (Term.Var x) =
  Polynomial.Poly [(Nat.one_nat, x)] (Util.Pos Nat.Zero_nat);
pleft alpha (Term.Fun f ts) =
  let {
    a = alpha f (List.size_list ts);
    (list, c) = a;
    (Polynomial.Poly lista cint) =
      Polynomial.polySum (Polynomial.Poly [] c)
        (List.foldra Polynomial.polySum
          (List.mapa
            (\ ct ->
              Polynomial.polyMult (fst ct) (Polynomial.pleft alpha (snd ct)))
            (List.zipa list ts))
          (Polynomial.Poly [] (Util.Pos Nat.Zero_nat)));
  } in (case lista of {
         [] -> (case cint of {
                 Util.Neg n -> Polynomial.Poly [] (Util.Pos Nat.Zero_nat);
                 Util.Pos nat -> Polynomial.Poly [] (Util.Pos nat);
               });
         aa : listaa -> Polynomial.Poly (aa : listaa) cint;
       });

pright ::
  forall a b.
    (Eq b) => (a -> Nat.Nat -> ([Nat.Nat], Util.Cint)) ->
                Term.Term a b -> Polynomial.Poly b;
pright alpha (Term.Var x) =
  Polynomial.Poly [(Nat.one_nat, x)] (Util.Pos Nat.Zero_nat);
pright alpha (Term.Fun f ts) =
  let {
    a = alpha f (List.size_list ts);
    (list, c) = a;
    (Polynomial.Poly vp aa) =
      Polynomial.polySum
        (List.foldra Polynomial.polySum
          (List.mapa
            (\ ct ->
              Polynomial.polyMult (fst ct) (Polynomial.pright alpha (snd ct)))
            (List.zipa list ts))
          (Polynomial.Poly [] (Util.Pos Nat.Zero_nat)))
        (Polynomial.Poly [] c);
  } in (case aa of {
         Util.Neg x -> Polynomial.Poly vp (Util.Pos Nat.Zero_nat);
         Util.Pos n -> Polynomial.Poly vp (Util.Pos n);
       });

genIndices :: forall a. Nat.Nat -> Integer -> [(Term.Term a [Char])];
genIndices Nat.Zero_nat uu = [];
genIndices (Nat.Suc n) i =
  Term.Var (List.append ['x', '_'] (Show.sshow_int i)) :
    Polynomial.genIndices n (i + 1);

genTerm :: forall a. a -> Nat.Nat -> Term.Term a [Char];
genTerm f n = Term.Fun f (Polynomial.genIndices n 1);

checkVPLists ::
  forall a. (Eq a) => [(Nat.Nat, a)] -> [(Nat.Nat, a)] -> Util.Eithera a ();
checkVPLists uu [] = ErrorMonad.returna ();
checkVPLists vp ((n, x) : vq) =
  (if n == Nat.Zero_nat then Polynomial.checkVPLists vp vq
    else (case Polynomial.lookupVar x vp of {
           Nothing -> ErrorMonad.errora x;
           Just a ->
             let {
               (m, vpp) = a;
             } in ErrorMonad.bind (ErrorMonad.checkB (Nat.less_eq_nat n m) x)
                    (\ xa -> Polynomial.checkVPLists vpp vq);
         }));

checkPoly ::
  forall a.
    (Eq a,
      Show.Sshow a) => Polynomial.Poly a ->
                         Polynomial.Poly a -> Util.Eithera [[Char]] ();
checkPoly (Polynomial.Poly vp c) (Polynomial.Poly vq d) =
  ErrorMonad.bind
    (ErrorMonad.checkB
      (case c of {
        Util.Neg x -> False;
        Util.Pos cn ->
          (case d of {
            Util.Neg x -> False;
            Util.Pos dn -> Nat.less_eq_nat dn cn;
          });
      })
      [['t', 'h', 'e', ' ', 'c', 'o', 'n', 's', 't', 'a', 'n', 't', ' ', 'o',
         'f', ' ', 't', 'h', 'e', ' ', 'l', 'e', 'f', 't', ' ', 'p', 'o', 'l',
         'y', 'n', 'o', 'm', 'i', 'a', 'l', ' ', 'i', 's', ' ', 't', 'o', 'o',
         ' ', 's', 'm', 'a', 'l', 'l']])
    (\ x ->
      ErrorMonad.catcha (Polynomial.checkVPLists vp vq)
        (\ xa ->
          ErrorMonad.errora
            [List.append
               (List.append ['v', 'a', 'r', 'i', 'a', 'b', 'l', 'e', ' ']
                 (Show.sshow xa))
               [' ', 'i', 's', ' ', 'p', 'r', 'o', 'b', 'l', 'e', 'm', 'a', 't',
                 'i', 'c']]));

create_af ::
  forall a.
    (Eq a) => [(a, (Util.Cint, [Nat.Nat]))] -> a -> Nat.Nat -> Nat.Nat -> Bool;
create_af [] f n i = True;
create_af ((g, (c, coeffs)) : rest) f n i =
  (if f == g && n == List.size_list coeffs
    then Nat.less_nat Nat.Zero_nat (List.nth coeffs i)
    else Polynomial.create_af rest f n i);

show_pair :: forall b. (Show.Sshow b) => (Nat.Nat, b) -> [Char];
show_pair (Nat.Zero_nat, v) = [];
show_pair (Nat.Suc Nat.Zero_nat, v) = Show.sshow v;
show_pair (Nat.Suc (Nat.Suc n), v) =
  List.append (List.append (Show.sshow_nat (Nat.Suc (Nat.Suc n))) ['*'])
    (Show.sshow v);

merge_entries :: [[Char]] -> [Char];
merge_entries [] = [];
merge_entries (x : xs) =
  List.append (List.append [' ', '+', ' '] x) (Polynomial.merge_entries xs);

show_number :: Util.Cint -> [Char];
show_number (Util.Neg Nat.Zero_nat) = [];
show_number (Util.Pos Nat.Zero_nat) = [];
show_number (Util.Neg (Nat.Suc n)) =
  List.append ['-'] (Show.sshow_nat (Nat.Suc n));
show_number (Util.Pos (Nat.Suc n)) = Show.sshow_nat (Nat.Suc n);

show_entries :: forall b. (Show.Sshow b) => Polynomial.Poly b -> [[Char]];
show_entries (Polynomial.Poly vp c) =
  List.filtera (\ entry -> not (List.nulla entry))
    (Polynomial.show_number c : List.mapa Polynomial.show_pair vp);

sshow_poly :: forall b. (Show.Sshow b) => Polynomial.Poly b -> [Char];
sshow_poly p =
  (case Polynomial.show_entries p of {
    [] -> ['0'];
    d : es -> List.append d (Polynomial.merge_entries es);
  });

addAtLeastOne :: forall a. Polynomial.Poly a -> Polynomial.Poly a;
addAtLeastOne (Polynomial.Poly vp (Util.Pos cn)) =
  Polynomial.Poly vp (Util.Pos (Nat.Suc cn));
addAtLeastOne (Polynomial.Poly vp (Util.Neg uu)) =
  Polynomial.Poly vp (Util.Pos (Nat.Suc Nat.Zero_nat));

checkPolyGt ::
  forall v.
    (Eq v,
      Show.Sshow v) => Polynomial.Poly v ->
                         Polynomial.Poly v -> Util.Eithera [[Char]] ();
checkPolyGt p q =
  ErrorMonad.catcha (Polynomial.checkPoly p (Polynomial.addAtLeastOne q))
    (\ x ->
      ErrorMonad.errora
        (List.concata
           [['C', 'o', 'u', 'l', 'd', ' ', 'n', 'o', 't', ' ', 'o', 'r', 'i',
              'e', 'n', 't', ' '],
             Polynomial.sshow_poly p, [' ', '>', ' '], Polynomial.sshow_poly q]
          : x));

checkSPolo ::
  forall f v.
    (Show.Sshow f, Eq v,
      Show.Sshow v) => (f -> Nat.Nat -> ([Nat.Nat], Util.Cint)) ->
                         (Term.Term f v, Term.Term f v) ->
                           Util.Eithera [[Char]] ();
checkSPolo alpha lr =
  ErrorMonad.catcha
    (Polynomial.checkPolyGt (Polynomial.pleft alpha (fst lr))
      (Polynomial.pright alpha (snd lr)))
    (\ x ->
      ErrorMonad.errora
        (List.concata
           [['C', 'o', 'u', 'l', 'd', ' ', 'n', 'o', 't', ' ', 'o', 'r', 'i',
              'e', 'n', 't', ' '],
             Term.sshow_term (fst lr), [' ', '>', ' '],
             Term.sshow_term (snd lr)]
          : x));

checkPolyGe ::
  forall v.
    (Eq v,
      Show.Sshow v) => Polynomial.Poly v ->
                         Polynomial.Poly v -> Util.Eithera [[Char]] ();
checkPolyGe p q =
  ErrorMonad.catcha (Polynomial.checkPoly p q)
    (\ x ->
      ErrorMonad.errora
        (List.concata
           [['C', 'o', 'u', 'l', 'd', ' ', 'n', 'o', 't', ' ', 'o', 'r', 'i',
              'e', 'n', 't', ' '],
             Polynomial.sshow_poly p, [' ', '>', '=', ' '],
             Polynomial.sshow_poly q]
          : x));

checkNSPolo ::
  forall f v.
    (Show.Sshow f, Eq v,
      Show.Sshow v) => (f -> Nat.Nat -> ([Nat.Nat], Util.Cint)) ->
                         (Term.Term f v, Term.Term f v) ->
                           Util.Eithera [[Char]] ();
checkNSPolo alpha lr =
  ErrorMonad.catcha
    (Polynomial.checkPolyGe (Polynomial.pleft alpha (fst lr))
      (Polynomial.pright alpha (snd lr)))
    (\ x ->
      ErrorMonad.errora
        (List.concata
           [['C', 'o', 'u', 'l', 'd', ' ', 'n', 'o', 't', ' ', 'o', 'r', 'i',
              'e', 'n', 't', ' '],
             Term.sshow_term (fst lr), [' ', '>', '=', ' '],
             Term.sshow_term (snd lr)]
          : x));

checkListPositive :: [Nat.Nat] -> Util.Eithera [[Char]] ();
checkListPositive [] = ErrorMonad.returna ();
checkListPositive (n : ns) =
  ErrorMonad.bind
    (ErrorMonad.checkB (Nat.less_nat Nat.Zero_nat n)
      [['C', 'o', 'e', 'f', 'f', 'i', 'c', 'i', 'e', 'n', 't', ' ', '0', ' ',
         '-', ' ', 'n', 'o', 't', ' ', 'm', 'o', 'n', 'o', 't', 'o', 'n', 'e']])
    (\ x -> Polynomial.checkListPositive ns);

checkPoloMono ::
  forall a. [(a, (Nat.Nat, [Nat.Nat]))] -> Util.Eithera [[Char]] ();
checkPoloMono [] = ErrorMonad.returna ();
checkPoloMono ((uu, (uv, coeffs)) : alpha) =
  ErrorMonad.bind (Polynomial.checkListPositive coeffs)
    (\ x -> Polynomial.checkPoloMono alpha);

showInterpretation ::
  forall a. (a -> Nat.Nat -> ([Nat.Nat], Util.Cint)) -> a -> Nat.Nat -> [Char];
showInterpretation alpha f n =
  Polynomial.sshow_poly (Polynomial.pleft alpha (Polynomial.genTerm f n));

createRepresentation ::
  forall a.
    (Eq a,
      Show.Sshow a) => (a -> Nat.Nat -> ([Nat.Nat], Util.Cint)) ->
                         [(a, Nat.Nat)] -> [[Char]];
createRepresentation alpha fs =
  ['p', 'o', 'l', 'y', 'n', 'o', 'm', 'i', 'a', 'l', ' ', 'i', 'n', 't', 'e',
    'r', 'p', 'r', 'e', 't', 'a', 't', 'i', 'o', 'n', ':', ' ']
    : List.mapa
        (\ (a @ (g, n)) ->
          List.concata
            [['P', 'o', 'l', '('], Show.sshow g, ['/'], Show.sshow_nat n,
              [')', ' ', '=', ' '], Polynomial.showInterpretation alpha g n])
        (List.remdups fs);

checkCintNonnegative :: Util.Cint -> Util.Eithera [[Char]] ();
checkCintNonnegative (Util.Pos uu) = ErrorMonad.returna ();
checkCintNonnegative (Util.Neg n) =
  ErrorMonad.checkB (n == Nat.Zero_nat)
    [['N', 'e', 'g', 'a', 't', 'i', 'v', 'e', ' ', 'c', 'o', 'n', 's', 't', 'a',
       'n', 't', '.']];

checkNegPoloMono ::
  forall a. [(a, (Util.Cint, [Nat.Nat]))] -> Util.Eithera [[Char]] ();
checkNegPoloMono [] = ErrorMonad.returna ();
checkNegPoloMono ((uu, (c, coeffs)) : alpha) =
  ErrorMonad.bind
    (ErrorMonad.bind (Polynomial.checkCintNonnegative c)
      (\ x -> Polynomial.checkListPositive coeffs))
    (\ x -> Polynomial.checkNegPoloMono alpha);

checkInterMono :: forall a. Polynomial.RedPair a -> Util.Eithera [[Char]] ();
checkInterMono (Polynomial.Polo alpha) = Polynomial.checkPoloMono alpha;
checkInterMono (Polynomial.NegPolo alphaa) = Polynomial.checkNegPoloMono alphaa;

redPairToPrePolyInter ::
  forall a. Polynomial.RedPair a -> [(a, (Util.Cint, [Nat.Nat]))];
redPairToPrePolyInter (Polynomial.Polo inter) =
  List.mapa
    (\ (a @ (f, aa)) ->
      let {
        (i, coeffs) = aa;
      } in (f, (Util.Pos i, coeffs)))
    inter;
redPairToPrePolyInter (Polynomial.NegPolo intera) = intera;

createPolyInter ::
  forall a.
    (Eq a) => [(a, (Util.Cint, [Nat.Nat]))] ->
                a -> Nat.Nat -> ([Nat.Nat], Util.Cint);
createPolyInter [] f n = (List.replicatea n Nat.one_nat, Util.intToCint 1);
createPolyInter ((g, (i, coeffs)) : rest) f n =
  (if f == g && n == List.size_list coeffs then (coeffs, i)
    else Polynomial.createPolyInter rest f n);

createRuleRemovalProc ::
  forall a b.
    (Eq a, Show.Sshow a, Eq b,
      Show.Sshow b) => Polynomial.RedPair a ->
                         [(Term.Term a b, Term.Term a b)] ->
                           [(Term.Term a b, Term.Term a b)] ->
                             Util.Eithera [[Char]] ();
createRuleRemovalProc interpr r rR =
  let {
    alpha =
      Polynomial.createPolyInter (Polynomial.redPairToPrePolyInter interpr);
  } in ErrorMonad.bind (Polynomial.checkInterMono interpr)
         (\ x ->
           RedPair.checkRuleRemovalProc (Polynomial.checkSPolo alpha)
             (Polynomial.checkNSPolo alpha)
             ['p', 'o', 'l', 'y', 'n', 'o', 'm', 'i', 'a', 'l', ' ', 'o', 'r',
               'd', 'e', 'r']
             (Polynomial.createRepresentation alpha
               (List.remdups
                 (List.mapa (\ (a @ (g, il)) -> (g, List.size_list (snd il)))
                   (Polynomial.redPairToPrePolyInter interpr))))
             r rR);

createMonoRedPairUrProc ::
  forall a b.
    (Eq a, Show.Sshow a, Eq b,
      Show.Sshow b) => Polynomial.RedPair a ->
                         [(Term.Term a b, Term.Term a b)] ->
                           [(Term.Term a b, Term.Term a b)] ->
                             [(Term.Term a b, Term.Term a b)] ->
                               [(Term.Term a b, Term.Term a b)] ->
                                 [(Term.Term a b, Term.Term a b)] ->
                                   Util.Eithera [[Char]] ();
createMonoRedPairUrProc interpr p rP r rR ur =
  let {
    alpha =
      Polynomial.createPolyInter (Polynomial.redPairToPrePolyInter interpr);
  } in ErrorMonad.bind (Polynomial.checkInterMono interpr)
         (\ x ->
           UsableRules.checkMonoRedPairUrProc (Polynomial.checkSPolo alpha)
             (Polynomial.checkNSPolo alpha)
             ['p', 'o', 'l', 'y', 'n', 'o', 'm', 'i', 'a', 'l', ' ', 'o', 'r',
               'd', 'e', 'r']
             (Polynomial.createRepresentation alpha
               (List.remdups
                 (List.mapa (\ (a @ (g, il)) -> (g, List.size_list (snd il)))
                   (Polynomial.redPairToPrePolyInter interpr))))
             p rP r rR ur);

createMonoRedPairProc ::
  forall a b.
    (Eq a, Show.Sshow a, Eq b,
      Show.Sshow b) => Polynomial.RedPair a ->
                         [(Term.Term a b, Term.Term a b)] ->
                           [(Term.Term a b, Term.Term a b)] ->
                             [(Term.Term a b, Term.Term a b)] ->
                               [(Term.Term a b, Term.Term a b)] ->
                                 Util.Eithera [[Char]] ();
createMonoRedPairProc interpr p rP r rR =
  let {
    alpha =
      Polynomial.createPolyInter (Polynomial.redPairToPrePolyInter interpr);
  } in ErrorMonad.bind (Polynomial.checkInterMono interpr)
         (\ x ->
           RedPair.checkMonoRedPairProc (Polynomial.checkSPolo alpha)
             (Polynomial.checkNSPolo alpha)
             ['p', 'o', 'l', 'y', 'n', 'o', 'm', 'i', 'a', 'l', ' ', 'o', 'r',
               'd', 'e', 'r']
             (Polynomial.createRepresentation alpha
               (List.remdups
                 (List.mapa (\ (a @ (g, il)) -> (g, List.size_list (snd il)))
                   (Polynomial.redPairToPrePolyInter interpr))))
             p rP r rR);

createRedPairUrAfProc ::
  forall a b.
    (Eq a, Show.Sshow a, Eq b,
      Show.Sshow b) => Polynomial.RedPair a ->
                         [(Term.Term a b, Term.Term a b)] ->
                           [(Term.Term a b, Term.Term a b)] ->
                             [(Term.Term a b, Term.Term a b)] ->
                               [(Term.Term a b, Term.Term a b)] ->
                                 Util.Eithera [[Char]] ();
createRedPairUrAfProc interpr p rP r ur =
  let {
    l = Polynomial.redPairToPrePolyInter interpr;
    alpha = Polynomial.createPolyInter l;
  } in UsableRules.checkRedPairUrAfProc (Polynomial.checkSPolo alpha)
         (Polynomial.checkNSPolo alpha)
         ['p', 'o', 'l', 'y', 'n', 'o', 'm', 'i', 'a', 'l', ' ', 'o', 'r', 'd',
           'e', 'r']
         (Polynomial.createRepresentation alpha
           (List.remdups
             (List.mapa (\ (a @ (g, il)) -> (g, List.size_list (snd il)))
               (Polynomial.redPairToPrePolyInter interpr))))
         p rP r ur (Polynomial.create_af l);

createRedPairProc ::
  forall a b.
    (Eq a, Show.Sshow a, Eq b,
      Show.Sshow b) => Polynomial.RedPair a ->
                         [(Term.Term a b, Term.Term a b)] ->
                           [(Term.Term a b, Term.Term a b)] ->
                             [(Term.Term a b, Term.Term a b)] ->
                               Util.Eithera [[Char]] ();
createRedPairProc interpr p rP r =
  let {
    alpha =
      Polynomial.createPolyInter (Polynomial.redPairToPrePolyInter interpr);
  } in RedPair.checkRedPairProc (Polynomial.checkSPolo alpha)
         (Polynomial.checkNSPolo alpha)
         ['p', 'o', 'l', 'y', 'n', 'o', 'm', 'i', 'a', 'l', ' ', 'o', 'r', 'd',
           'e', 'r']
         (Polynomial.createRepresentation alpha
           (List.remdups
             (List.mapa (\ (a @ (g, il)) -> (g, List.size_list (snd il)))
               (Polynomial.redPairToPrePolyInter interpr))))
         p rP r;

}
