(*
Copyright 2009 Christian Sternagel, René Thiemann, Sarah Winkler, Harald Zankl

This file is part of IsaFoR/CeTA.

IsaFoR/CeTA is free software: you can redistribute it and/or modify it under the
terms of the GNU Lesser General Public License as published by the Free Software
Foundation, either version 3 of the License, or (at your option) any later
version.

IsaFoR/CeTA is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License along
with IsaFoR/CeTA. If not, see <http://www.gnu.org/licenses/>.
*)
theory RedPair imports Trs DpFramework begin


definition one_step_ctxt_closed :: "('f,'v)trel \<Rightarrow> bool"
where
  "one_step_ctxt_closed r = (\<forall> f bef s t aft. ((s,t) \<in> r \<longrightarrow> (Fun f (bef @ s # aft), Fun f (bef @ t # aft)) \<in> r))"

lemma one_step_ctxt_closed_imp_ctxt_closed: "one_step_ctxt_closed r \<Longrightarrow> ctxt_closed r"
proof (unfold one_step_ctxt_closed_def ctxt_closed_def)
  assume one: "(\<forall> f bef s t aft. ((s,t) \<in> r \<longrightarrow> (Fun f (bef @ s # aft), Fun f (bef @ t # aft)) \<in> r))"
  show "CC r \<subseteq> r"
  proof
    fix csct
    assume csct: "csct \<in> CC r"
    thus "csct \<in> r"
    proof (cases csct)
      case (Pair cs ct)
      with csct have "(cs, ct) \<in> CC r" by simp
      thus ?thesis
      proof
	fix s t C 
	assume "cs = C\<langle>s\<rangle>" and "ct = C\<langle>t\<rangle>" and st: "(s,t) \<in> r"
	from st have "(C\<langle>s\<rangle>,C\<langle>t\<rangle>) \<in> r" 
	proof (induct C arbitrary: s t)
	  case Hole
	  thus ?case by simp
	next
	  case (More f bef C aft)
	  with one show ?case by auto
	qed
	with prems show ?thesis by simp 
      qed
    qed
  qed
qed

lemma ctxt_closed_imp_one_step_ctxt_closed : fixes r :: "('f,'v)trel"
  assumes "ctxt_closed r"
  shows "one_step_ctxt_closed r"
proof (unfold one_step_ctxt_closed_def, clarify)
    fix f :: 'f and  bef and s t :: "('f,'v)term" and aft :: "('f,'v) term list"
    assume step: "(s,t) \<in> r"
    let ?C = "More f bef \<box> aft"
    have "Fun f (bef @ s # aft) = ?C\<langle>s\<rangle>" (is "?l = _")
    and  "Fun f (bef @ t # aft) = ?C\<langle>t\<rangle>" (is "?r = _") by auto
    hence "(?l,?r) \<in> CC r" using step ccI[where s = s and t = t and r = r and C = ?C] by simp
    thus "(?l,?r) \<in> r" using assms unfolding ctxt_closed_def by auto
qed

lemma one_step_ctxt_closed_rstep: "one_step_ctxt_closed (rstep r)"
using ctxt_closed_imp_one_step_ctxt_closed ctxt_closed_rstep by auto


definition all_ctxt_closed :: "('f,'v)trel \<Rightarrow> bool"
where "all_ctxt_closed rel \<equiv> \<forall> f ts ss. (length ts = length ss \<and> (\<forall> i. i < length ts \<longrightarrow> (ts ! i, ss ! i) \<in> rel) \<longrightarrow> (Fun f ts, Fun f ss) \<in> rel)"


lemma tr_one_imp_all_ctxt_closed: assumes tran: "trans r" 
  and refl: "refl r"
  and one: "one_step_ctxt_closed r" 
  shows "all_ctxt_closed r"
unfolding all_ctxt_closed_def
proof (clarify)
  fix f ts ss
  assume l: "length ts = length ss" and
         steps: "\<forall> i < length ts. (ts ! i, ss ! i) \<in> r"
  let ?n = "length ts"
  let ?us = "\<lambda> i. take i ss @ (take (?n - i) (List.drop i ts))"
  have "\<And> i. i \<le> ?n \<Longrightarrow> (Fun f ts, Fun f (?us i)) \<in> r"
  proof -
    fix i
    assume "i \<le> ?n"
    thus "(Fun f ts, Fun f (?us i)) \<in> r"
    proof (induct i)
      case 0 thus ?case using refl unfolding refl_on_def by auto
    next
      case (Suc i)
      hence i: "i < length ts" by auto
      from Suc have ind: "(Fun f ts, Fun f (?us i)) \<in> r" by auto
      from Suc steps have step: "(ts ! i, ss ! i) \<in> r" by auto
      let ?bef = "take i ss"
      let ?aft = "take (?n - Suc i) (List.drop (Suc i) ts)"
      from one step have cstep: "(Fun f (?bef @ ts ! i # ?aft), Fun f (?bef @ ss ! i # ?aft)) \<in> r" (is "(?l,?r) \<in> r") unfolding one_step_ctxt_closed_def by auto
      have lId: "Fun f (?us i) = ?l" using l i by (auto simp: drop_Suc_conv_tl)
      have rId: "Fun f (?us (Suc i)) = ?r" using l i by (auto simp: take_Suc_conv_app_nth)
      from cstep have "(Fun f (?us i), Fun f (?us (Suc i))) \<in> r" by (simp only: lId rId)
      with  ind show ?case using tran unfolding trans_def by best
    qed
  qed      
  hence "(Fun f ts, Fun f (?us ?n)) \<in> r" by force
  thus "(Fun f ts, Fun f ss) \<in> r" using l by  auto
qed


lemma rtrancl_one_step_ctxt_closed : assumes 
  "one_step_ctxt_closed r" 
  shows "one_step_ctxt_closed (r^*)"
unfolding one_step_ctxt_closed_def 
proof clarify
  fix f bef s t aft
  assume "(s,t) \<in> r^*"
  thus "(Fun f (bef @ s # aft), Fun f (bef @ t # aft)) \<in> r^*"
  proof (induct rule: rtrancl_induct, simp)
    fix t u
    assume a: "(t,u) \<in> r" and
      b: "(Fun f (bef @ s # aft), Fun f (bef @ t # aft)) \<in> r^*"
    from a assms have "(Fun f (bef @ t # aft), Fun f (bef @ u # aft)) \<in> r" unfolding one_step_ctxt_closed_def by auto
    with b show "(Fun f (bef @ s # aft), Fun f (bef @ u # aft)) \<in> r^*" by auto
  qed
qed



lemma rstep_rtrancl_all_ctxt_closed: 
  shows "all_ctxt_closed ((rstep r)^*)"
proof -
  have a: "one_step_ctxt_closed ((rstep r)^*)"
    by (rule rtrancl_one_step_ctxt_closed, auto simp: one_step_ctxt_closed_rstep)
  show ?thesis 
    by (rule tr_one_imp_all_ctxt_closed, auto simp: trans_rtrancl refl_rtrancl a)
qed


lemma R_inclusion : assumes C: "one_step_ctxt_closed Rel" 
  and s: "sub_closed Rel"
  and orient: "R \<subseteq> Rel" 
  shows "rstep R \<subseteq> Rel"
proof -
  from C have D: "ctxt_closed Rel" by (rule one_step_ctxt_closed_imp_ctxt_closed)
  from orient have "CS R \<subseteq> CS Rel" by (rule CS_monotone)
  with s have "CS R \<subseteq> Rel" by (auto simp: sub_closed_def)
  hence "rstep R \<subseteq> CC Rel" by (rule CC_monotone)
  with D show ?thesis by (auto simp: ctxt_closed_def)
qed
      

lemma rstepSubsetMono:
 assumes "rR \<subseteq> R"
 shows "rstep rR \<subseteq> rstep R"
proof-
 from assms have "CS rR \<subseteq> CS R" using CS_monotone by auto
 thus "CC (CS rR) \<subseteq> CC (CS R)" using CC_monotone by auto
qed


locale redpair =    
  fixes  S :: "('f,'v)trel" ("\<succ>")
    and   NS :: "('f,'v)trel" ("\<succeq>")
  assumes SN: "SN \<succ>"
    and compat: "\<succ> O \<succeq> \<subseteq> \<succ>"
    and wmNS: "one_step_ctxt_closed \<succeq>"
    and subNS: "sub_closed \<succeq>"
    and subS: "sub_closed \<succ>"
begin

lemma mNS: "ctxt_closed \<succeq>" 
  using wmNS by (rule one_step_ctxt_closed_imp_ctxt_closed) 

lemma trCompat : "\<succ> O \<succeq>^* \<subseteq> \<succ>" using compat by (rule compat_tr_compat)


lemma trans_split_NS_S_union: (* SW *)
  assumes "(rstep R)^* \<subseteq> (\<succeq> \<union> \<succ>)^*"
  shows "(rstep R)^* \<subseteq> \<succeq>^* O \<succ>^* O \<succ> \<union> \<succeq>^*"
proof(rule subsetI2)
  fix x y
  assume "(x, y)  \<in> (rstep R)^*"
  with assms have  "(x,y) \<in> (\<succeq> \<union> \<succ>)^*" by auto
  thus "(x,y) \<in> \<succeq>^* O \<succ>^* O \<succ> \<union> \<succeq>^*" using split_NS_S_union[where S = S and NS = NS] compat by auto
qed

  
lemma redPairSound :
  assumes rules: "R \<subseteq> \<succeq>"
  and nonStrict: "rP \<inter> P \<subseteq> \<succeq> \<union> \<succ>"
  and strict: "P - rP \<subseteq> \<succ>"
  and chain: "ichain (P,R) s t \<sigma>"
  shows  "\<exists> j. ichain (rP, R) (shift s j) (shift t j) (shift \<sigma> j)"
using chain
proof auto
  let ?ss = "\<lambda> i. s i \<cdot> \<sigma> i"
  let ?ts = "\<lambda> i. t i \<cdot> \<sigma> i"
  let ?next = "\<lambda> i. (s (Suc i), t (Suc i))"
  assume allP: "\<forall> i. (s i, t i) \<in> P"
  assume steps: "\<forall> i. (?ts i, ?ss (Suc i)) \<in> (rstep R)^*"
  have piece1: "\<forall> i. (?ts i, ?ts (Suc i)) \<in> \<succ> \<or> (?ts i, ?ts (Suc i)) \<in> \<succeq>^* \<and> (?next i) \<in> rP"
  proof
    fix i
    from subNS wmNS rules R_inclusion have "rstep R \<subseteq> \<succeq>" by auto
    hence "(rstep R)^* \<subseteq> \<succeq>^*" by (rule rtrancl_mono)
    with steps have stepsNS: "(?ts i, ?ss (Suc i)) \<in> \<succeq>^*" by blast
    show "(?ts i, ?ts (Suc i)) \<in> \<succ> \<or> (?ts i, ?ts (Suc i)) \<in> \<succeq>^* \<and> (?next i) \<in> rP" 
    proof (cases "?next i \<in> \<succ> \<inter> (P \<union> rP)")
      case True
      hence "(?ss (Suc i), ?ts (Suc i)) \<in> \<succ>" using subS by (auto simp: sub_closed_def)
      with stepsNS trCompat show ?thesis by auto 
    next
      case False
      with strict nonStrict allP
      have one: "?next i \<in> \<succeq>" and two: "?next i \<in> rP" by auto
      from one have "(?ss (Suc i), ?ts (Suc i)) \<in> \<succeq>" using subNS by (auto simp: sub_closed_def)
      with stepsNS have "(?ts i, ?ts (Suc i)) \<in> \<succeq>^*" by auto
      with two show ?thesis by simp
    qed
  qed
  
  hence infSeq: "\<forall> i. (?ts i, ?ts (Suc i)) \<in> \<succeq>^* \<union> \<succ>" by auto
  from SN have "SN \<succ> (?ts 0)" by auto
  from infSeq trCompat this have "\<exists> j. \<forall> i \<ge> j. (?ts i, ?ts (Suc i)) \<in> \<succeq>^* - \<succ>" by (rule non_strict_ending)
  
  from this obtain j where "\<forall> i \<ge> j. (?ts i, ?ts (Suc i)) \<in> \<succeq>^* - \<succ>" ..
  with piece1 have "\<forall> i \<ge> j. ?next i \<in> rP" by blast
  hence "\<forall> i. (s (i+(Suc j)), t (i + (Suc j))) \<in> rP" by auto
  thus "\<exists> j. \<forall> i. (s (i+j), t (i+j)) \<in> rP" ..
qed

lemma redPairProcessorDpp : 
  assumes rules: "R \<subseteq> \<succeq>"
  and nonStrict: "rP \<inter> P \<subseteq> \<succeq> \<union> \<succ>"
  and strict: "P - rP \<subseteq> \<succ>"
  and finite: "finite_dpp (rP, R)"
  shows "finite_dpp (P,R)"
proof (unfold finite_dpp_def, rule notI, (erule exE)+)
  fix s t \<sigma>
  assume min: "min_ichain (P,R) s t \<sigma>"
  hence "ichain (P,R) s t \<sigma>" by simp
  with rules nonStrict strict have "\<exists> j. ichain (rP, R) (shift s j) (shift t j) (shift \<sigma> j)" by (rule redPairSound)
  from this obtain j where "ichain (rP, R) (shift s j) (shift t j) (shift \<sigma> j)" ..
  with min have "min_ichain (rP, R) (shift s j) (shift t j) (shift \<sigma> j)" by simp
  with finite show False by (unfold finite_dpp_def, blast)
qed
 
lemma redPairProcessorGdpp : 
  assumes rules: "R \<subseteq> \<succeq>"
  and nonStrict: "rP \<inter> nodes P \<subseteq> \<succeq> \<union> \<succ>"
  and strict: "nodes P - rP \<subseteq> \<succ>"
  and finite: "finite_gdpp (node_restriction P rP, R)" (is "finite_gdpp (?rP, R)")
  shows "finite_gdpp (P,R)"
proof -
  have "\<And> s t \<sigma>. min_ichain (nodes P, R) s t \<sigma> \<Longrightarrow> \<exists> j. min_ichain (rP, R) (shift s j) (shift t j) (shift \<sigma> j)"
  proof - 
    fix s t \<sigma>
    assume min: "min_ichain (nodes P,R) s t \<sigma>"
    hence "ichain (nodes P,R) s t \<sigma>" by simp
    with rules nonStrict strict have "\<exists> j. ichain (rP, R) (shift s j) (shift t j) (shift \<sigma> j)" by (rule redPairSound)
    from this obtain j where "ichain (rP, R) (shift s j) (shift t j) (shift \<sigma> j)" ..
    with min have "min_ichain (rP, R) (shift s j) (shift t j) (shift \<sigma> j)" by simp
    thus "\<exists> j. min_ichain (rP, R) (shift s j) (shift t j) (shift \<sigma> j)" ..
  qed
  from this and finite show ?thesis by (rule dpp_lift_to_gdpp)
qed

lemma monoRedPairSound : (* SW *)
  assumes monotonic: "one_step_ctxt_closed \<succ>"
  and rNonStrict: "rR \<inter> R \<subseteq> \<succeq> \<union> \<succ>"
  and rStrict: "R - rR \<subseteq> \<succ>"
  and pNonStrict: "rP \<inter> P \<subseteq> \<succeq> \<union> \<succ>"
  and pStrict: "P - rP \<subseteq> \<succ>"
  and chain: "ichain (P,R) s t \<sigma>"
  shows  "\<exists> j. ichain (rP, rR) (shift s j) (shift t j) (shift \<sigma> j)"
using chain
proof auto
  let ?ss = "\<lambda> i. s i \<cdot> \<sigma> i"
  let ?ts = "\<lambda> i. t i \<cdot> \<sigma> i"
  let ?nextr = "\<lambda> i. (?ts i, ?ss (Suc i))"
  let ?nextp = "\<lambda> i. (s i, t i)"
  assume allP: "\<forall> i. (s i, t i) \<in> P"
  assume steps: "\<forall> i. (?ts i, ?ss (Suc i)) \<in> (rstep R)^*"
  from monotonic wmNS have monoNSS:"one_step_ctxt_closed (\<succeq> \<union> \<succ>)" unfolding one_step_ctxt_closed_def by auto
  from subS subNS have subNSS:"sub_closed (\<succeq> \<union> \<succ>)" unfolding sub_closed_def using cs_union_distr[where ?R="\<succeq>" and ?S=\<succ>] by auto
  let ?SS = "\<succeq>^* O \<succ>^* O \<succ>"
  from rStrict rNonStrict have "R \<subseteq> \<succeq> \<union> \<succ>" by auto
  with monoNSS subNSS R_inclusion have "(rstep R) \<subseteq> (\<succeq> \<union> \<succ>)" by auto
  hence "(rstep R)^* \<subseteq> (\<succeq> \<union> \<succ>)^*" by (rule rtrancl_mono)
  hence rsteps_in_NSS_or_NS:"(rstep R)^* \<subseteq> ?SS \<union> \<succeq>^*" using trans_split_NS_S_union by auto
  have piece1: "\<forall> i. (?ss i, ?ss (Suc i)) \<in> ?SS \<or> (?ss i, ?ss (Suc i)) \<in> \<succeq>^* \<and> ((?nextr i) \<in> (rstep rR)^* \<and> (?nextp i) \<in> rP)"
  proof
    fix i
    show "(?ss i, ?ss (Suc i)) \<in> ?SS \<or> (?ss i, ?ss (Suc i)) \<in> \<succeq>^* \<and> (?nextr i) \<in> (rstep rR)^* \<and> (?nextp i) \<in> rP"
    proof (cases "?nextp i \<in> \<succ> \<inter> (P \<union> rP)")
      (* case 1: strict decrease with DP step *)
      case True
      hence "(?ss i, ?ts i) \<in> \<succ>" using subS by (auto simp: sub_closed_def)
      from steps rsteps_in_NSS_or_NS have "(?nextr i) \<in> ?SS \<union> \<succeq>^*" by auto
      with `(?ss i, ?ts i) \<in> \<succ>` have "(?ss i, ?ss (Suc i)) \<in> (?SS \<union> \<succeq>^*) O \<succ>" by auto
      hence "(?ss i, ?ss (Suc i)) \<in> ?SS" by simp
      thus ?thesis by auto
    next
      case False
      with pStrict pNonStrict allP
      have "?nextp i \<in> \<succeq>" and "?nextp i \<in> rP" by auto
      hence dp_step_in_NS: "(?ss i, ?ts i) \<in> \<succeq>" using subNS by (auto simp: sub_closed_def)
      show "(?ss i, ?ss (Suc i)) \<in> ?SS \<or> (?ss i, ?ss (Suc i)) \<in> \<succeq>^* \<and> (?nextr i) \<in> (rstep rR)^* \<and> (?nextp i) \<in> rP"
      proof (cases "(?nextr i) \<in> ((rstep rR)^* - ?SS)")
        (* case 2: strict decrease in R steps *)
        case False
        hence or:"(?nextr i) \<notin> ((rstep rR)^*) \<or> (?nextr i) \<in> ?SS" by auto
        have "(?nextr i) \<in> ?SS"
        proof (cases "(?nextr i) \<in> ?SS", simp)
          case False
          with or steps have "(?nextr i) \<in> ((rstep R)^* ) - ((rstep rR)^* )" by auto
          hence "(?nextr i) \<in> (rstep R)^* O ((rstep R) - (rstep rR)) O (rstep R)^*"
            using rtrancl_diff_decomp[where ?A="rstep R"] by auto
          hence "\<exists> u v. (?ts i, u) \<in> (rstep R)^* \<and> (u, v) \<in> ((rstep R) - (rstep rR)) \<and> (v, ?ss (Suc i)) \<in> (rstep R)^*" by auto
          then obtain u v where
            tu:"(?ts i, u) \<in> (rstep R)^*" and uv:"(u, v) \<in> ((rstep R) - (rstep rR))" and vs:"(v, ?ss (Suc i)) \<in> (rstep R)^*" by auto
          with rsteps_in_NSS_or_NS have tu:"(?ts i, u) \<in> ?SS \<union> \<succeq>^*" and vs:"(v, ?ss (Suc i)) \<in> ?SS \<union> \<succeq>^*" by auto
          from rStrict monotonic subS R_inclusion have "rstep (R - rR) \<subseteq> \<succ>" by auto
          with uv have "(u, v) \<in> \<succ>" using rstep_diff by auto
          with tu have "(?ts i, v) \<in> (\<succ> O ?SS) \<union> (\<succ> O \<succeq>^*)" by auto
          with trCompat have "(?ts i, v) \<in> (\<succ> O  \<succeq>^* O \<succ>^* O \<succ>) \<union> \<succ>" by auto
          with trCompat have "(?ts i, v) \<in> (\<succ> O \<succ>^* O \<succ>) \<union> \<succ>" by auto
          hence "(?ts i, v) \<in> \<succ>^* O \<succ>" by auto
          with vs have nextrstep:"(?ts i, ?ss (Suc i)) \<in> (\<succeq>^* O \<succ>^* O \<succ> O \<succ>^* O \<succ>) \<union> (\<succeq>^* O \<succ>^* O \<succ>)" by auto
          have "\<succ>^* O \<succ> O \<succ>^* O \<succ> \<subseteq> (\<succ>^* O \<succ>^*) O \<succ>" using O_assoc[where ?R="\<succ>^*" and ?S="\<succ>^*" and ?T="\<succ>"] by auto
          hence "\<succ>^* O \<succ> O \<succ>^* O \<succ> \<subseteq> \<succ>^* O \<succ>" using rtrancl_idemp_self_comp by auto
          hence "\<succeq>^* O \<succ>^* O \<succ> O \<succ>^* O \<succ> \<subseteq> \<succeq>^* O \<succ>^* O \<succ>"
            using rel_comp_mono[where ?r'="\<succeq>^*" and ?r="\<succeq>^*" and ?s'="\<succ>^* O \<succ> O \<succ>^* O \<succ>" and ?s="\<succ>^* O \<succ>"] by auto
          hence "(\<succeq>^* O \<succ>^* O \<succ> O \<succ>^* O \<succ>) \<union> (\<succeq>^* O \<succ>^* O \<succ>) \<subseteq> (\<succeq>^* O \<succ>^* O \<succ>)" by auto
          from this nextrstep show  "(?ts i, ?ss (Suc i)) \<in> \<succeq>^* O \<succ>^* O \<succ>" by (rule set_mp)
        qed
        with dp_step_in_NS trCompat show ?thesis by auto
      next
        (* case 3: only weak decrease with both DP and R steps *)
        case True
        hence "(?nextr i) \<in> ((rstep rR)^*)" by auto
        from steps rsteps_in_NSS_or_NS True have "(?nextr i) \<in> \<succeq>^*" by auto
        from dp_step_in_NS `(?nextr i) \<in> \<succeq>^*` have  "(?ss i, ?ss (Suc i)) \<in> \<succeq>^*" by auto
        with `?nextp i \<in> rP` `(?nextr i) \<in> ((rstep rR)^*)` show ?thesis by simp
      qed
    qed
  qed
  hence infSeq: "\<forall> i. (?ss i, ?ss (Suc i)) \<in>  \<succeq>^* \<union> (\<succeq>^* O \<succ>^* O \<succ>)" by auto
  from trCompat have compat2:"(\<succeq>^* O \<succ>^* O \<succ>) O \<succeq>^*  \<subseteq> (\<succeq>^* O \<succ>^* O \<succ>)" by auto
  have "SN (\<succeq>^* O \<succ>^* O \<succ>)" using NS_comp_S_is_SN[where S = S and NS = NS] SN compat by force
  hence "SN (\<succeq>^* O \<succ>^* O \<succ>) (?ss 0)" by auto
  from infSeq compat2 this have "\<exists> j. \<forall> i \<ge> j. (?ss i, ?ss (Suc i)) \<in> \<succeq>^* - (\<succeq>^* O \<succ>^* O \<succ>)" using non_strict_ending by auto
  from this obtain j where "\<forall> i \<ge> j. (?ss i, ?ss (Suc i)) \<in> \<succeq>^* - (\<succeq>^* O \<succ>^* O \<succ>)" ..
  with piece1 have "\<forall> i \<ge> j. (?nextr i \<in> ((rstep rR)^*) \<and> (?nextp i) \<in> rP)" by blast
  hence "\<forall> i. ((?ts (i + j), ?ss (Suc i + j)) \<in> ((rstep rR)^*)  \<and> (s (i + j), t (i + j)) \<in> rP)" by auto
  thus "\<exists>j. (\<forall>i. (s (i + j), t (i + j)) \<in> rP) \<and> (\<forall>i. (t (i + j) \<cdot> \<sigma> (i + j), s (Suc (i + j)) \<cdot> \<sigma> (Suc (i + j))) \<in> (rstep rR)\<^sup>*)" by auto
qed

(*
lemma finiteReducedDpp :
 assumes "rR \<subseteq> R" and "rP \<subseteq> P" and "finite_dpp (P,R)"
 shows "finite_dpp (rP, rR)"
proof (unfold finite_dpp_def, rule notI, (erule exE)+)
  fix s t \<sigma>
  assume min: "min_ichain (rP,rR) s t \<sigma>"
  hence "ichain (rP,rR) s t \<sigma>" by simp
(*  with monotonic rNonStrict rStrict pNonStrict pStrict have "\<exists> j. ichain (rP, rR) (shift s j) (shift t j) (shift \<sigma> j)"
   by (rule monoRedPairSound)
  then obtain j where rchain:"ichain (rP, rR) (shift s j) (shift t j) (shift \<sigma> j)" by auto*)
  from `rR \<subseteq> R` have "CS rR \<subseteq> CS R" by (rule CS_monotone)
  hence "rstep rR \<subseteq> rstep R" by (rule CC_monotone)
  from min have alli:"(\<forall>i. SN (rstep rR) (s i \<cdot> \<sigma> i)) \<and> (\<forall>i. SN (rstep rR) (t i \<cdot> \<sigma> i))" by simp
  have "(\<forall>i. SN (rstep R) (s i \<cdot> \<sigma> i) \<and> SN (rstep R) (t i \<cdot> \<sigma> i))"
  proof
   fix i
   from alli have "SN (rstep rR) (s i \<cdot> \<sigma> i)" and "SN (rstep rR) (t i \<cdot> \<sigma> i)" by auto
   with  `rstep rR \<subseteq> rstep R` have "SN (rstep rR) (s i \<cdot> \<sigma> i)" and "SN (rstep rR) (t i \<cdot> \<sigma> i)"
    using SN_imp_SN_subset[where ?R'="rstep R" and ?rR'="rstep rR"] by auto
   thus " SN (rstep rR) (s i \<cdot> \<sigma> i) \<and> SN (rstep rR) (t i \<cdot> \<sigma> i)" by auto
  qed
  hence "(\<forall>i. SN (rstep rR) (s i \<cdot> \<sigma> i)) \<and> (\<forall>i. SN (rstep rR) (t i \<cdot> \<sigma> i))" by (auto simp:all_conj_distrib)
  with rchain have "min_ichain (rP, rR) (shift s j) (shift t j) (shift \<sigma> j)" using min_ichain.simps by auto
*)

lemma monoRedPairProcessorDpp :
  assumes monotonic: "one_step_ctxt_closed \<succ>"
  and rNonStrict: "rR \<inter> R \<subseteq> \<succeq> \<union> S"
  and rStrict: "R - rR \<subseteq> S"
  and pNonStrict: "rP \<inter> P \<subseteq> \<succeq> \<union> S"
  and pStrict: "P - rP \<subseteq> S"
  and finite: "finite_dpp (rP, rR)"
  and "rR \<subseteq> R"
  shows "finite_dpp (P,R)"
proof (unfold finite_dpp_def, rule notI, (erule exE)+)
  fix s t \<sigma>
  assume min: "min_ichain (P,R) s t \<sigma>"
  hence "ichain (P,R) s t \<sigma>" by simp
  with monotonic rNonStrict rStrict pNonStrict pStrict have "\<exists> j. ichain (rP, rR) (shift s j) (shift t j) (shift \<sigma> j)"
    by (rule monoRedPairSound)
  then obtain j where rchain:"ichain (rP, rR) (shift s j) (shift t j) (shift \<sigma> j)" by auto
  from `rR \<subseteq> R` have "CS rR \<subseteq> CS R" by (rule CS_monotone)
  hence "rstep rR \<subseteq> rstep R" by (rule CC_monotone)
  from min have alli:"(\<forall>i. SN (rstep R) (s i \<cdot> \<sigma> i)) \<and> (\<forall>i. SN (rstep R) (t i \<cdot> \<sigma> i))" by simp
  have "(\<forall>i. SN (rstep rR) (s i \<cdot> \<sigma> i) \<and> SN (rstep rR) (t i \<cdot> \<sigma> i))"
  proof
    fix i
    from alli have "SN (rstep R) (s i \<cdot> \<sigma> i)" and "SN (rstep R) (t i \<cdot> \<sigma> i)" by auto
    with  `rstep rR \<subseteq> rstep R` have "SN (rstep rR) (s i \<cdot> \<sigma> i)" and "SN (rstep rR) (t i \<cdot> \<sigma> i)"
      using SN_imp_SN_subset[where ?R'="rstep R" and ?rR'="rstep rR"] by auto
    thus " SN (rstep rR) (s i \<cdot> \<sigma> i) \<and> SN (rstep rR) (t i \<cdot> \<sigma> i)" by auto
  qed
  hence "(\<forall>i. SN (rstep rR) (s i \<cdot> \<sigma> i)) \<and> (\<forall>i. SN (rstep rR) (t i \<cdot> \<sigma> i))" by (auto simp:all_conj_distrib)
  with rchain have "min_ichain (rP, rR) (shift s j) (shift t j) (shift \<sigma> j)" using min_ichain.simps by auto
  with finite show False by (unfold finite_dpp_def, blast)
qed


lemma ruleRemoval:
  assumes monotonic: "one_step_ctxt_closed \<succ>"
  and nonStrict: "R \<subseteq> \<succeq> \<union> \<succ>"
  and strict: "R - rR \<subseteq> \<succ>"
  and subset:"rR \<subseteq> R"
  shows "((SN :: ('f, 'v)trel \<Rightarrow> bool)(rstep R)) = (SN (rstep rR))"
proof
  assume "SN (rstep R)"
  show "SN (rstep rR)"
  proof
    fix  x :: "('f, 'v) term"
    from subset have "rstep rR \<subseteq> rstep R" using rstepSubsetMono by auto
    from `SN (rstep R)` have "SN (rstep R) x" by auto
    from this `rstep rR \<subseteq> rstep R` show "SN (rstep rR) x" by (rule SN_imp_SN_subset)
  qed
next
  assume "SN (rstep rR)"
  show "SN (rstep R)"
  proof
    fix x :: "('f, 'v) term"
    show "SN (rstep R) x"
    proof(rule ccontr)
      assume "\<not> SN (rstep R) x"
      hence "\<exists> f. f 0 = x \<and> (\<forall>i. (f i, f (Suc i)) \<in> rstep R)" using SN_elt_def by auto
      then obtain f where "f 0 = x" and alli:"\<forall>i. (f i, f (Suc i)) \<in> rstep R" by auto
      from SN have "SN \<succ> (f 0)" by auto
      from monotonic strict subS R_inclusion have "(rstep (R - rR)) \<subseteq> \<succ>" by auto
      from monotonic wmNS have monoNSS:"one_step_ctxt_closed (\<succeq> \<union> \<succ>)" unfolding one_step_ctxt_closed_def by auto
      from subS subNS have subNSS:"sub_closed (\<succeq> \<union> \<succ>)" unfolding sub_closed_def using cs_union_distr[where ?R="\<succeq>" and ?S=\<succ>] by auto
      with nonStrict monoNSS subNSS R_inclusion have "(rstep R) \<subseteq> (\<succeq> \<union> \<succ>)" by auto
      with `\<forall>i. (f i, f (Suc i)) \<in> rstep R` have "\<forall>i. (f i, f (Suc i)) \<in> (\<succeq> \<union> \<succ>)" by auto
      with compat `SN \<succ> (f 0)` have "\<exists> j. \<forall> i \<ge> j.  (f i, f (Suc i)) \<in> (\<succeq> - \<succ>)" using non_strict_ending by auto
      then obtain j where stepNS:" \<forall> i \<ge> j.  (f i, f (Suc i)) \<in> (\<succeq> - \<succ>)" by auto
      have alli:"\<forall> i \<ge> j.  (f i, f (Suc i)) \<in> (rstep rR)"
      proof
	fix i
	show "i \<ge> j \<longrightarrow> (f i, f (Suc i)) \<in> rstep rR"
	proof
	  assume "i \<ge> j"
	  from alli have "(f i, f (Suc i)) \<in> rstep R" by auto thm rstep_diff
	  with `i \<ge> j` stepNS `(rstep (R - rR)) \<subseteq> \<succ>` have "(f i, f (Suc i)) \<in>  (rstep R - rstep (R - rR))" by auto
	  hence "(f i, f (Suc i)) \<in>  rstep (R - (R - rR))" using rstep_diff[where ?R=R and ?R'="R - rR"] by auto
	  from `rR \<subseteq> R` have "R - (R - rR) \<subseteq> rR" by auto
	  with `(f i, f (Suc i)) \<in>  rstep (R - (R - rR))`  show "(f i, f (Suc i)) \<in> (rstep rR)"  using rstepSubsetMono by auto
	qed
      qed
      let ?g = "\<lambda> k. f (k + j)"
      from alli have "\<forall> k. (?g k, ?g(Suc k)) \<in> (rstep rR)" and "?g 0 = f j" by auto
      hence "\<exists>g. g 0 = (f j) \<and> (\<forall>k. (g k, g (Suc k)) \<in> (rstep rR))" by best
      with `SN (rstep rR)` show False by auto
    qed
  qed
qed

end

(* now the implementation starts *)

(*
  fixes checkS :: "('f :: sshow, 'v :: sshow)rule \<Rightarrow> sstring list monad"
    and checkNS :: "('f :: sshow, 'v :: sshow)rule \<Rightarrow> sstring list monad"
    and name :: "sstring" 
    and representation :: "sstring list"
  assumes checkSSound: "isOK(checkS rule) \<Longrightarrow> rule \<in> S" 
    and   checkNSSound: "isOK(checkNS rule)  \<Longrightarrow> rule \<in> NS"
begin
*)


fun checkRedPairProc :: "
  (('f :: sshow,'v :: sshow)rule \<Rightarrow> sstring list monad) \<Rightarrow>
  (('f,'v)rule \<Rightarrow> sstring list monad) \<Rightarrow> 
   sstring \<Rightarrow>
   sstring list \<Rightarrow> 
   ('f,'v)trsL \<Rightarrow> ('f,'v)trsL \<Rightarrow> ('f,'v)trsL  \<Rightarrow> sstring list monad"
where "checkRedPairProc checkSS checkNSS  name representation P rP r  = 
( let (Sp :: ('f,'v) trsL ,NSp) = computeSplit (\<lambda> foo. isOK(checkSS foo)) P in
    (
   checkAllM checkNSS r
    \<leftarrow> (\<lambda> s. ''problem when orienting TRS'' # s ) 
 \<guillemotright> checkAllM checkNSS NSp
    \<leftarrow> (\<lambda> s. ''problem when orienting DPs'' # s )
 \<guillemotright> checkSubset NSp rP
    \<leftarrow> (\<lambda> s. case checkSS s of Left err \<Rightarrow> err | Right ok \<Rightarrow> [''internal error'', ''please contact one of the CeTA members'',''http://cl-informatik.uibk.ac.at/software/ceta'',''ceta@informatik.uibk.ac.at''])
    )
  \<leftarrow> (\<lambda> s. (''could not apply the reduction pair processor with the following''
         # (representation @ ([] # s))))
 )"

fun checkMonoRedPairProc :: "
  (('f :: sshow,'v :: sshow)rule \<Rightarrow> sstring list monad) \<Rightarrow>
  (('f,'v)rule \<Rightarrow> sstring list monad) \<Rightarrow> 
   sstring \<Rightarrow>
   sstring list \<Rightarrow> 
   ('f,'v)trsL \<Rightarrow> ('f,'v)trsL \<Rightarrow> ('f,'v)trsL  \<Rightarrow>  ('f,'v)trsL  \<Rightarrow> sstring list monad"
where "checkMonoRedPairProc checkSS checkNSS  name representation P rP R rR  = 
( let (Sp :: ('f,'v) trsL , NSp) = computeSplit (\<lambda> foo. isOK(checkSS foo)) P in
  let (Sr :: ('f,'v) trsL , NSr) = computeSplit (\<lambda> foo. isOK(checkSS foo)) R in
    (
   checkAllM checkNSS NSr
    \<leftarrow> (\<lambda> s. ''problem when orienting TRS'' # s ) 
 \<guillemotright> checkAllM checkNSS NSp
    \<leftarrow> (\<lambda> s. ''problem when orienting DPs'' # s )
 \<guillemotright> checkSubset rR R
    \<leftarrow> (\<lambda> s. [''problem: additional ''])
 \<guillemotright> checkSubset NSr rR
    \<leftarrow> (\<lambda> s. case checkSS s of Left err \<Rightarrow> err | Right ok \<Rightarrow> [''internal error'', ''please contact one of the CeTA members'',''http://cl-informatik.uibk.ac.at/software/ceta'',''ceta@informatik.uibk.ac.at''])
 \<guillemotright> checkSubset NSp rP
    \<leftarrow> (\<lambda> s. case checkSS s of Left err \<Rightarrow> err | Right ok \<Rightarrow> [''internal error'', ''please contact one of the CeTA members'',''http://cl-informatik.uibk.ac.at/software/ceta'',''ceta@informatik.uibk.ac.at''])
    )
  \<leftarrow> (\<lambda> s. (''could not apply the reduction pair processor with the following''
         # (representation @ ([] # s))))
 )"

fun checkRuleRemovalProc :: "
  (('f :: sshow,'v :: sshow)rule \<Rightarrow> sstring list monad) \<Rightarrow>
  (('f,'v)rule \<Rightarrow> sstring list monad) \<Rightarrow> 
   sstring \<Rightarrow>
   sstring list \<Rightarrow> 
   ('f,'v)trsL \<Rightarrow> ('f,'v)trsL \<Rightarrow> sstring list monad"
where "checkRuleRemovalProc checkSS checkNSS  name representation R rR  = 
( let (Sp :: ('f,'v) trsL , NSp) = computeSplit (\<lambda> foo. isOK(checkSS foo)) R in
    (
   checkAllM checkNSS R
    \<leftarrow> (\<lambda> s. ''problem when orienting rules'' # s ) 
 \<guillemotright> checkSubset NSp rR
    \<leftarrow> (\<lambda> s. case checkSS s of Left err \<Rightarrow> err | Right ok \<Rightarrow> [''internal error'', ''please contact one of the CeTA members'',''http://cl-informatik.uibk.ac.at/software/ceta'',''ceta@informatik.uibk.ac.at''])
 \<guillemotright> checkSubset rR R
    \<leftarrow> (\<lambda> s. case checkSS s of Left err \<Rightarrow> err | Right ok \<Rightarrow> [''internal error'', ''please contact one of the CeTA members'',''http://cl-informatik.uibk.ac.at/software/ceta'',''ceta@informatik.uibk.ac.at''])
    )
  \<leftarrow> (\<lambda> s. (''could not apply the rule removal processor with the following''
         # (representation @ ([] # s))))
 )"
end
